package method.dictionary;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

public class SearchSeedTerms {

	/**
	 * configuration file of domain representation
	 */
	String configure;
	
	public SearchSeedTerms(String file) {
		configure = file;
	}
	
	/**
	 * search and output all seed terms given the domain representation
	 * @throws IOException
	 * @throws SAXException
	 * @throws ParserConfigurationException
	 */
	public void search() throws IOException, SAXException, ParserConfigurationException {
		DomainConfig config = new DomainConfig();
		config.readConfigFile(configure);
		
		Hashtable<String, HashSet<String>> termFromTypes = searchSemanticTypeDescendant(config.getClassTypeMapping());
		Hashtable<String, HashSet<String>> termFromConcepts = searchConceptDescendant(config.getClassConceptMapping());
		
		Hashtable<String, HashSet<String>> terms = new Hashtable<String,HashSet<String>>();
		
		for(String entityClass : termFromTypes.keySet()) {
			terms.put(entityClass, termFromTypes.get(entityClass));
		}
		for(String entityClass : termFromConcepts.keySet()) {
			if(terms.containsKey(entityClass)) {
				terms.get(entityClass).addAll(termFromConcepts.get(entityClass));
			}
			else {
				terms.put(entityClass, termFromConcepts.get(entityClass));
			}
		}
		
		print(terms);
	}
	
	/**
	 * the function extracts seed terms from UMLS using the semantic types assigned to each entity class
	 * @param class_type_mapping, the mapping from entity class to semantic types
	 * @return the mapping from entity class to seed terms
	 * @throws IOException
	 */
	private Hashtable<String, HashSet<String>> searchSemanticTypeDescendant(Hashtable<String, HashSet<String>> class_type_mapping) throws IOException {
		Hashtable<String, HashSet<String>> concepts = new Hashtable<String, HashSet<String>>();
		Hashtable<String, HashSet<String>> terms = new Hashtable<String, HashSet<String>>();
		
		for(String entityClass : class_type_mapping.keySet()) {
			concepts.put(entityClass, new HashSet<String>());
			terms.put(entityClass, new HashSet<String>());
		}
		
		//collect all concepts belonging to the semantic types 
		for(BufferedReader reader = new BufferedReader(new FileReader("MRSTY.RRF")); reader.ready();) {
			String[] fields = reader.readLine().split("\\|");
			String conceptID = fields[0];
			String type = fields[3];
			for(String entityClass : class_type_mapping.keySet()) {
				if(class_type_mapping.get(entityClass).contains(type))
					concepts.get(entityClass).add(conceptID);
			}
		}
		
		//collect all terms
		for(BufferedReader reader = new BufferedReader(new FileReader("MRCONSO.RRF")); reader.ready();){			
			String[] fields = reader.readLine().split("\\|");
			String term = fields[14];
			String conceptID = fields[0];
			for(String entityClass : concepts.keySet()) {
				if(concepts.get(entityClass).contains(conceptID))
					terms.get(entityClass).add(term);
			}
		}
		return terms;
	}
	
	/**
	 * the function extracts seed terms from UMLS using the UMLS concepts assigned to each entity class
	 * @param class_concept_mapping, the mapping from entity class to UMLS concepts
	 * @return the mapping from entity class to seed terms
	 * @throws IOException
	 */
	private Hashtable<String, HashSet<String>> searchConceptDescendant(Hashtable<String, HashSet<String>> class_concept_mapping) throws IOException {
		Hashtable<String, HashSet<String>> relationship = new Hashtable<String, HashSet<String>>();
		Hashtable<String, HashSet<String>> concepts = new Hashtable<String, HashSet<String>>();
		Hashtable<String, HashSet<String>> terms = new Hashtable<String, HashSet<String>>();
		
		//extract all qualified relationships from UMLS metathesaurus
		for (BufferedReader reader = new BufferedReader(new FileReader("MRREL.RRF")); reader.ready();) {
			String line = reader.readLine();
			String[] fields = line.split("\\|");

			if(!fields[3].equals("CHD"))
				continue;
			if(relationship.containsKey(fields[0])) {
				relationship.get(fields[0]).add(fields[4]);
			}
			else {
				relationship.put(fields[0], new HashSet<String>());
				relationship.get(fields[0]).add(fields[4]);
			}
		}
		
		//collect all descendent concepts for an entity class.
		for(String entityClass: class_concept_mapping.keySet()) {
			concepts.put(entityClass, new HashSet<String>());
			terms.put(entityClass, new HashSet<String>());
			for(String concept : class_concept_mapping.get(entityClass)) {
				HashSet<String> descendents = new HashSet<String>();
				descCal(concept, descendents, relationship);
				concepts.get(entityClass).addAll(descendents);
			}
		}
		
		//collect all terms
		for(BufferedReader reader = new BufferedReader(new FileReader("MRCONSO.RRF")); reader.ready();){			
			String[] fields = reader.readLine().split("\\|");
			String term = fields[14];
			String conceptID = fields[0];
			for(String entityClass : concepts.keySet()) {
				if(concepts.get(entityClass).contains(conceptID))
					terms.get(entityClass).add(term);
			}
		}
		return terms;
	}
	
	
	/**
	 * recursively get descendents
	 * @param ancestor
	 * @param descendents
	 * @param relationship
	 * @throws IOException
	 */
	private void descCal(String ancestor, HashSet<String> descendents, Hashtable<String, HashSet<String>> relationship) throws IOException {
		HashSet<String> result = new HashSet<String>();
		if(!relationship.containsKey(ancestor))
			return;
		Iterator<String> ite = relationship.get(ancestor).iterator();
		while(ite.hasNext()) {
			String son = ite.next();
			if(son.equals(ancestor))
				continue;
			if(!descendents.contains(son))
				result.add(son);
				descendents.add(son);
		}
		if(result.size() == 0)
			return;
		
		ite = result.iterator();
		while(ite.hasNext()) {
			descCal(ite.next(), descendents, relationship);
		}
	}
	

	private void print(Hashtable<String, HashSet<String>> terms) throws IOException {
		BufferedWriter writer = null;
		
		File dir = new File("seedterm");
		if(!dir.isDirectory())
			dir.mkdir();
		
		for(String entityClass : terms.keySet()) {
			writer = new BufferedWriter(new FileWriter("seedterm/" + entityClass + ".desc"));
			for(String desc : terms.get(entityClass)) {
				writer.write(desc + "\n");
			}
			writer.flush();
		}
		
		writer.close();
	}
	
}

