package method.signature;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;

import opennlp.tools.chunker.ChunkerME;
import opennlp.tools.chunker.ChunkerModel;
import opennlp.tools.postag.POSModel;
import opennlp.tools.postag.POSTaggerME;

import util.RegularExpression;

public class SignatureSimilarity{
	//seed terms
	private LinkedList<HashSet<String>> seed = new LinkedList<HashSet<String>>();
	
	//mapping from index to entity class
	private Hashtable<String, String> type = new Hashtable<String, String>();
	
	//stopword list
	private HashSet<String> stopwords;

	//similarity threshold for entity classification, default value 0.002
	private double threshold = 0.002;
	
	//how many context words to include before and after
	private int context = 2; 
	
	//weight of inner words over context words
	private double inner = 20;	
	
	//idf table for all words
	private Hashtable<String, Double> IDF = new Hashtable<String, Double>();
	
	//signature vectors for all entity classes
	private LinkedList<Hashtable<String, Double>> signatures;
	
	//for the sake of similarity computing
	private LinkedList<Double> norms;
	
	//backup corpus for signature generation, could also be the test corpus itself.
	private String backupCorpus;
	
	//the signature generator
	private SignatureGenerator sigGen;
	
	private boolean regenerate;
	
	/**
	 * 
	 * @param corpus, backup corpus (could also be the test corpus itself) for signature calculation. 
	 * 		original texts needed. each line should be a sentence.
	 * @param context
	 * @param threshold
	 * @param inner
	 * @param regenrate, true if you want to regenerate the signatures. 
	 * false if you don't AND you already have signatures generated (stored in the folder ).
	 * @throws IOException
	 */
	public SignatureSimilarity(String corpus, int context, double threshold, double inner, boolean reg) throws IOException {
		signatures = new LinkedList<Hashtable<String, Double>>();
		backupCorpus = corpus;
		this.context = context;
		this.threshold = threshold;
		this.inner = inner;
		regenerate = reg;
		norms = new LinkedList<Double>();
		
		loadSeedTerms();

		idfCalculation();
		
		sigGen = new SignatureGenerator(IDF);

		furtherProcessDicTerms();
	}
	
	/**
	 * load seed terms and stop words
	 * @throws IOException
	 */
	private void loadSeedTerms() throws IOException {
		BufferedReader reader;
		File dir = new File("seedterm");
		stopwords = new HashSet<String>();
		reader = new BufferedReader(new FileReader("stopwords.txt"));
		while(reader.ready()) {
			stopwords.add(reader.readLine());
		}
		
		if(!dir.isDirectory()) {
			System.err.println("no directory for seed terms");
		}
		
		String[] fileList = dir.list();
		for(int i = 0; i < fileList.length; ++ i) {
			reader = new BufferedReader(new FileReader("seedterm/" + fileList[i]));
			HashSet<String> dec = new HashSet<String> ();
			
			type.put(Integer.toString(i), fileList[i].substring(0, fileList[i].length() - 5));
			
			while(reader.ready()) {
				dec.add(reader.readLine());
			}
			seed.add(dec);
		}
	}
	
	/**
	 * calculate the idf table for all words
	 * @throws IOException
	 */
	private void idfCalculation() throws IOException {
		double total = 0;
		for(BufferedReader sr = new BufferedReader(new FileReader(backupCorpus)); sr.ready(); ) {
			String line = sr.readLine().replace('\t', ' ').toLowerCase();
			String[] words = line.split(" ");
			HashSet<String> wordSet = new HashSet<String>();
			for(String word : words) {
				if(RegularExpression.isNumber(word))
					word = "DIGIT";
				wordSet.add(word);
			}
			for(String word : wordSet) {
				if(IDF.containsKey(word)) {
					IDF.put(word, IDF.get(word) + 1);
				}
				else {
					IDF.put(word, 1.);
				}
			}
			total ++;
		}
		for(String word : IDF.keySet()) {
			IDF.put(word, Math.log(total/IDF.get(word)));
		}
	}
	
	/**
	 * 	generate combined signature for each entity class
	 */
	private void furtherProcessDicTerms() throws IOException {
		File dir = new File("signatures");
		dir.mkdir();
		for(int i = 0; i < seed.size(); ++ i) {
			File file = new File("signatures/" + type.get(Integer.toString(i)) + ".txt");
			Hashtable<String, Double> tmp = new Hashtable<String, Double>();
			
			if(file.isFile() && !regenerate) {
				BufferedReader reader = new BufferedReader(new FileReader(file));
				while(reader.ready()) {
					String line = reader.readLine();
					String[] tokens = line.split("\t");
					tmp.put(tokens[0], Double.parseDouble(tokens[1]));
				}
			}
			else {
				tmp = sigGen.generateFrequencySignatureForSet(seed.get(i), backupCorpus, context, inner, true);
				BufferedWriter writer = new BufferedWriter(new FileWriter(file));
				for(String key : tmp.keySet()){
					writer.write(key + "\t" + tmp.get(key) +"\n");
				}
				writer.flush();
				writer.close();
			}
			
			signatures.add(tmp);			
			double norm = 0;
			for(String word : tmp.keySet()){
				Double aa = tmp.get(word);
				norm += aa*aa;
			}
			norms.add(Math.sqrt(norm));
		}
	}
	
	/**
	 * compute similarity between a term and an entity type
	 * @param begin: starting position of the term in the sentence
	 * @param end: ending position of the term in the sentence
	 * @param sentence: the sentence from which the term comes
	 * @param type: index of the entity class being compared
	 * @return
	 * @throws IOException
	 */
	private double similarity(int begin, int end, String[] sentence, int type) throws IOException {
		if(begin == end && stopwords.contains(sentence[begin]))
			return 0.0;
		Hashtable<String, Double> sig = sigGen.generateSignatureForTerm(begin, end, sentence, context, inner);
		
		double score = computeSimilarity(sig, type);

		if(Double.isNaN(score))
			return 0.0;
		return score;
	}
	
	/**
	 * compute the cosine similarity
	 * @param a: a signature vector
	 * @param type: index of the entity class being compared.
	 * @return
	 */
	private double computeSimilarity(Hashtable<String, Double> a, int type) {
		//This is optimized for the case that b is significantly larger than a
		Enumeration<String> ite = a.keys();
		double norm1 = 0;
		double dotPro = 0;
		double norm2 = norms.get(type);
		while(ite.hasMoreElements()) {
			String word = ite.nextElement();
			Double aa = a.get(word);
			norm1 += aa*aa;
			if(signatures.get(type).containsKey(word)) {
				dotPro += aa * signatures.get(type).get(word);
			}
		}
		norm1 = Math.sqrt(norm1);
		
		return dotPro / norm1 / norm2;
	}
	
	/**
	 * main function for named entity recognition
	 * @param testfile, the input file of test data
	 * @throws IOException
	 */
	public void recognize(String testfile) throws IOException {
		InputStream modelIn = new FileInputStream("en-pos-maxent.bin");
		POSTaggerME tagger = new POSTaggerME(new POSModel(modelIn));
		
		modelIn = new FileInputStream("en-chunker.bin");
		ChunkerME chunker = new ChunkerME(new ChunkerModel(modelIn));
				
		BufferedReader reader = new BufferedReader(new FileReader(testfile));
		BufferedWriter writer = new BufferedWriter(new FileWriter(testfile + "_chunked"));
		
		System.out.println("chunking text...");
		
		while(reader.ready()) {
			String line = reader.readLine();
			String[] tokens = line.split(" ");
			String[] pos = tagger.tag(tokens);
			String[] chunks = chunker.chunk(tokens, pos);
			
			for(int i = 0; i < tokens.length; ++ i) {
				writer.write(tokens[i] + "\t" + chunks[i] + "\n");
			}
			writer.write("\n");
		}
		
		writer.flush();
		writer.close();
		writer = new BufferedWriter(new FileWriter(testfile.substring(0, testfile.length())+ "_tagged"));
		LinkedList<String> lines_for_sentence = new LinkedList<String>();
		
		System.out.println("recognizing ne...");
		
		for (reader = new BufferedReader(new FileReader(testfile + "_chunked")); reader.ready(); ) {
			String line = reader.readLine();
			if(line.length() == 0) {
				//process last sentence
				String[] words = new String[lines_for_sentence.size()];
				String[] chunks = new String[lines_for_sentence.size()];
				
				for(int i = 0; i < lines_for_sentence.size(); ++ i) {
					String[] splitted_line =  lines_for_sentence.get(i).split("\t");
					words[i] = splitted_line[0].toLowerCase();
					chunks[i] = splitted_line[1];			
				}

				String[] labels = new String[lines_for_sentence.size()];
				
				for(int i = 0 ;i < labels.length; ++ i) {
					labels[i] = "O";
				}
				
				for(int begin = 0; begin < labels.length; ++ begin) {
					for(int end = begin; end < labels.length; ++ end) {
						//Series of conditions which control candidates
						if(chunks[begin].startsWith("B-NP")) {
							boolean flag = true;
							for(int kk = begin + 1; kk <= end; ++ kk) {
								if(!chunks[kk].equals("I-NP")) {
									flag = false;
									break;
								}
							}
							if(!flag){
								continue;
							}
							if(end != words.length - 1)
								if(chunks[end + 1].equals("I-NP"))
									continue;
							
							/*
							 * filter candidates by IDF
							 */
							double idf = 0;
							for(int m = begin; m <= end; ++ m) {
								Double get = IDF.get(words[m]);
								if(get != null)
									idf += IDF.get(words[m]);
							}
							idf = idf/(end - begin + 1);
							if(idf < 4 || idf > 8)
								continue;
							
							/*
							 * Only consider entire NP chunks
							 */
							//initializing sets
							double[] scores = new double[seed.size()];
							
							for(int m = 0; m < seed.size(); ++ m) {
								scores[m] = similarity(begin, end, words, m);
							}
							
							//select label;
							String label = "";
							for(int m = 0; m < scores.length; ++ m) {
								flag = true;
								for(int n = 0; n < scores.length; ++ n) {
									if(scores[m] < scores[n]) {
										flag = false;
										break;
									}
								}
								if(flag && scores[m] > threshold) {
									label = type.get(Integer.toString(m));
									break;
								}
							}
							
							if(label.length() > 1) { 
								labels[begin] = "B-" + label;
								if(end != begin){
									for(int j = begin + 1; j <= end ; ++ j) {
										labels[j] = "I-" + label;
									}
								}
							}
						}
					}
				}
							
									
				/*
				 * output results
				 */
				for(int i = 0; i < lines_for_sentence.size(); ++ i) {
					String _line = lines_for_sentence.get(i);
					writer.write(_line + "\t" + labels[i] + "\n");
				}
				writer.write("\n");
				writer.flush();
				lines_for_sentence = new LinkedList<String>();
				continue;
			}
			lines_for_sentence.add(line);
		}
		writer.flush();
		writer.close();
	}
	
}
